/*
 * midifile 1.20
 * 
 * Read a MIDI file.
 * Major rewrite to make it more specific for mtm
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "midifile.h"
#include "mtm.h"

/* public stuff */

long Mf_currtime = 0L;		/* current time in delta-time units     */
int Mf_nomerge = 0;		/* 1 => continued system exclusives are */
			        /* not collapsed.                       */
long discards;			/* count of discarded note starts       */
/* private stuff */
#define MSGINCREMENT 128
static char *Msgbuff = NULL;	/* message buffer */
static int Msgsize = 0;		/* Size of currently allocated Msg */
static int Msgindex = 0;	/* index of next available location in Msg */
static long Mf_toberead = 0L;
static int ntrks;		/* number of tracks provided */

// debugging aid
extern long bytecount;		// maintained by filegetc in mtm.c
extern unsigned char debugbuff[];
static int bytecounter=-1;

static void mferror(char *s) {
  mtm_error(s);
  exit(1);
}

static int egetc(void) {	/* read a single character and abort on EOF */
  int c = mtm_getc();		/* get character routine in mtm.c */
  
  bytecounter++;
  if (c == EOF)
    mferror("premature EOF");
  Mf_toberead--;
  return(c);
}

static void chanmessage(int status, int c1, int c2) {
  int chan = status & 0xf;

  switch (status & 0xf0) {
  case 0x80:
    mtm_noteoff(chan,c1,c2);
    break;
  case 0x90:
    mtm_noteon(chan,c1,c2);
    break;
  case 0xa0:
    mtm_pressure(chan,c1,c2);
    break;
  case 0xb0:
    mtm_parameter(chan,c1,c2);
    break;
  case 0xc0:
    mtm_program(chan,c1);
    break;
  case 0xd0:
    mtm_chanpressure(chan,c1);
    break;
  case 0xe0:
    mtm_pitchbend(chan,c1,c2);
    break;
  }
}

/* The code below allows collection of a system exclusive message of */
/* arbitrary length.  The Msgbuff is expanded as necessary.  The only */
/* visible data/routines are msginit(), msgadd(), msg(), msgleng(). */

static void msginit(void) {
  Msgindex = 0;
}

static char *msg(void) {
  return(Msgbuff);
}

static int msgleng(void) {
  return(Msgindex);
}

static void biggermsg(void) {
  char *newmess;
  char *oldmess = Msgbuff;
  int i;
  int oldleng = Msgsize;

  Msgsize += MSGINCREMENT;
  newmess = (char *) malloc( (unsigned)(sizeof(char)*Msgsize) );

  if (newmess == NULL)
    mferror("malloc error!");
		
  /* copy old message into larger new one */
  if ( oldmess != NULL ) {
    for (i=0; i<oldleng; i++)
      newmess[i] = oldmess[i];
    free(oldmess);
  }
  Msgbuff = newmess;
}

static void msgadd(int c) {
  /* If necessary, allocate larger message buffer. */
  if (Msgindex >= Msgsize)
    biggermsg();
  Msgbuff[Msgindex++] = c;
}

static void sysex(void) {
  mtm_sysex(msgleng(),msg());
}

/* readvarinum - read a varying-length number, and return the */
/* number of characters it took. */
static long readvarinum(void) {
  long value;
  int c;

  c = egetc();
  value = c & 0x7f;
  while (c & 0x80) {
    c = egetc();
    value = (value << 7) + (c & 0x7f);
  }
  return (value);
}

static long to32bit(int c1, int c2, int c3, int c4) {
  long value = 0L;

  value = (c1 & 0xff);
  value = (value<<8) + (c2 & 0xff);
  value = (value<<8) + (c3 & 0xff);
  value = (value<<8) + (c4 & 0xff);
  return (value);
}

static int to16bit(int c1, int c2) {
  return ((c1 & 0xff ) << 8) + (c2 & 0xff);
}

static long read32bit(void) {
  int c1, c2, c3, c4;

  c1 = egetc();
  c2 = egetc();
  c3 = egetc();
  c4 = egetc();
  return to32bit(c1,c2,c3,c4);
}

static int read16bit(void) {
  int c1, c2;
  c1 = egetc();
  c2 = egetc();
  return to16bit(c1,c2);
}

static int readmt(char *s) {	/* read through the "MThd" or "MTrk" header string */
  int n = 0;
  int c;
  char *p = s;
  char buff[32];

  while ( n++<4 && (c=mtm_getc()) != EOF ) {
    if ( c != *p++ ) {
      (void) strcpy(buff,"expecting ");
      (void) strcat(buff,s);
      mferror(buff);
    }
  }
  return(c);
}

static void readheader(void) {		/* read a header chunk */
  int format, division;

  if ( readmt("MThd") == EOF )
    return;

  Mf_toberead = read32bit();
  format = read16bit();
  ntrks = read16bit();
  division = read16bit();

  if (format == 0)
    ntrks=17;				// 16 + trk 0
  mtm_header(format, ntrks, division);	/* process the track header */
  /* flush any extra stuff, in case the length of header is not 6 */
  while ( Mf_toberead > 0 )
    (void) egetc();
}

static void badbyte(int c) {
  char buff[128], *p;
  int i, off;

  (void) sprintf(buff,"unexpected byte: 0x%02x at position %ld\n", c, bytecount);
  off = strlen(buff);
  p = buff + off;
  off = bytecount + 1;
  for (i=0; i<16; i++) {
    sprintf(p, "%02x ", debugbuff[(off+i)&15]);
    p += 3;
  }
  strcpy(p, "\n");
  mferror(buff);
}

static void metaevent(int type) {
  int leng = msgleng();
  char *m = msg();

  switch  ( type ) {
  case 0x00:
    mtm_seqnum(to16bit(m[0],m[1]));
    break;
  case 0x01:	/* Text event */
  case 0x02:	/* Copyright notice */
  case 0x03:	/* Sequence/Track name */
  case 0x04:	/* Instrument name */
  case 0x05:	/* Lyric */
  case 0x06:	/* Marker */
  case 0x07:	/* Cue point */
  case 0x08:
  case 0x09:
  case 0x0a:
  case 0x0b:
  case 0x0c:
  case 0x0d:
  case 0x0e:
  case 0x0f:
    /* These are all text events */
    mtm_text(type,leng,m);
    break;
  case 0x2f:	/* End of Track */
    mtm_eot();
    break;
  case 0x51:	/* Set tempo */
    mtm_tempo(to32bit(0,m[0],m[1],m[2]));
    break;
  case 0x54:
    mtm_smpte(m[0],m[1],m[2],m[3],m[4]);
    break;
  case 0x58:
    mtm_timesig(m[0],m[1],m[2],m[3]);
    break;
  case 0x59:
    mtm_keysig((signed)m[0],m[1]);
    break;
  case 0x7f:
    mtm_seqspecific(m[0], leng, m);
    break;
  default:
    mtm_metamisc(type,leng,m);
  }
}

static int readtrack(void) {		 /* read a track chunk */
  /* This array is indexed by the high half of a status byte.  It's */
  /* value is either the number of bytes needed (1 or 2) for a channel */
  /* message, or 0 (meaning it's not  a channel message). */
  static int chantype[] = {
    0, 0, 0, 0, 0, 0, 0, 0,		/* 0x00 through 0x70 */
    2, 2, 2, 2, 1, 1, 2, 0		/* 0x80 through 0xf0 */
  };
  long lookfor;
  int i, c, c1, type;
  int sysexcontinue = 0;	/* 1 if last message was an unfinished sysex */
  int running = 0;	/* 1 when running status used */
  int status = 0;		/* status value (e.g. 0x90==note-on) */
  int needed;

  if ( readmt("MTrk") == EOF )
    return(0);

  Mf_toberead = read32bit();
  Mf_currtime = 0;

  mtm_trackstart();
  while ( Mf_toberead > 0 ) {
    Mf_currtime += readvarinum();	/* delta time */
    c = egetc();

    if (sysexcontinue && c != 0xf7)
      mferror("didn't find expected continuation of a sysex");
    if ((c & 0xf0) == 0xf0) {
      // handle system messages
      switch ( c ) {
      case 0xff:			/* meta event */
	type = egetc();
	lookfor = readvarinum();
	msginit();
	for (i=0; i<lookfor; i++)
	  msgadd(egetc());
	metaevent(type);
	break;
	
      case 0xf0:		/* start of system exclusive */
	lookfor = readvarinum();
	msginit();
	msgadd(0xf0);
	for (i=0; i<lookfor; i++)
	  msgadd(c=egetc());
	if ( c==0xf7 || Mf_nomerge==0 )
	  sysex();
	else
	  sysexcontinue = 1;  /* merge into next msg */
	break;
	
      case 0xf7:	/* sysex continuation or arbitrary stuff */
	lookfor = readvarinum();
	if ( ! sysexcontinue )
	  msginit();
	for (i=0; i<lookfor; i++)
	  msgadd(c=egetc());
	if (! sysexcontinue) {
	  mtm_arbitrary(msgleng(),msg());
	}
	else if ( c == 0xf7 ) {
	  sysex();
	  sysexcontinue = 0;
	}
	break;
      case 0xfe:		// this shouldn't really occur in a midifile
	break;		// so we just discard it...
      default:
	badbyte(c);
	break;
      }
    }
    else {
      if ((c & 0x80) == 0) {	 /* running status? */
	if ( status == 0 )
	  mferror("unexpected running status");
	running = 1;
      }
      else {
	status = c;
	running = 0;
      }
      needed = chantype[ (status>>4) & 0xf ];
      if (needed) {		/* ie. is it a channel message? */
	if (running)
	  c1 = c;
	else
	  c1 = egetc();
	chanmessage(status, c1, (needed>1) ? egetc() : 0 );
      }
    }
  }
  mtm_trackend();
  return(1);
}

void mfread(void) { 	/* The only non-static function in this file. */
  int i;

  discards = 0;
  readheader();
  for (i=0; i<ntrks; i++)
    readtrack();
  i = mtm_getc();
  if (i != EOF) 
    fprintf(stderr, "mtm: garbage at end of midifile ignored\n");
  if (discards != 0)
    fprintf(stderr, "mtm: %ld duplicated note discarded\n", discards);
}
